// Generated from Flocc program
// see http://www.flocc.net/

#ifndef HEADER_H
#define HEADER_H

#include <mpi.h>

//#define DEBUG 1
#include "pointer.h"
//#include "hashing.h"
//#include "vecmaps.h"
//#include "reparter.h"
#include "carttops.h"

// BASE TYPES 

typedef struct nullStruct {
  bool operator<(const nullStruct& b) const {
    return false;
  }
  bool operator==(const nullStruct& b) const {
    return true;
  }
} nullStruct;

// PRINTING

void printVal(int v) { printf("%i", v); }
void printVal(const char *v) { printf("%s", v); }
void printVal(nullStruct v) {
  printVal("null");
}
void printVal(int* v) {
  // only safe to print one, as don't know array length
  printVal(*v); 
} 

// MPI STUFF

// converts the rank of a process in one communicator to it's rank
// in another communicator.
void translateRank(MPI::Comm *srcComm, int srcRank, MPI::Comm *destComm, int *destRank) {
  MPI_Group orig_group, new_group; 
  orig_group = srcComm->Get_group();
  new_group = destComm->Get_group();
  MPI::Group::Translate_ranks(orig_group, 1, &srcRank, new_group, destRank);
} 

//typedef void MPI::Comm::Errhandler_fn(MPI::Comm &, int *, ... );
void throwErrHandler(MPI::Comm &comm, int *errCode, ...) {
  int len = 5000;
  char str[len];
  for (int i = 0; i < len; i++) str[i] = 0;
  MPI::Get_error_string(*errCode, str, len);
  std::cerr << "MPI Error: " << str << "\n";
  throw 1;
}

// TIME

#define MAX_TIME_HEADER 1024
int timeRank;
char timeHeader[MAX_TIME_HEADER];
double beginTime, middleTime;

void initTime() {
  // rank
  timeRank = MPI::COMM_WORLD.Get_rank();

  // start time
  beginTime = MPI_Wtime();
  middleTime = beginTime;
}

void logTime(const char* event, double *startTime) {
  if (timeRank == 0) {
    double End = MPI_Wtime();
    std::cerr << timeHeader << "\t" 
              << event << "\t" << (End - (*startTime)) << std::endl;
  }
  middleTime = MPI_Wtime();
}

#endif

// Global cartesian communicator
static const int ndims = 1;
static const int maxNodesPerDim = 4096;
static std::vector<int> dimSizes;
static MPI::Cartcomm cartComm;
static int coordRanks[maxNodesPerDim];
static int thisCoord[ndims];
static int thisRank, Nproc, rootRank;


 // user def reduction function
  void userFun98(const void* invec, void *inoutvec, int len, const MPI::Datatype& datatype) {
   const double* in = (const double*)invec;
   double* out = (double*)inoutvec;
   // TODO: WHEN SENDING PROPER DATATYPE, CHANGE (len / sizeof(double)) to len!
   for (int i = 0; i < (len / sizeof(double)); i++) {
     double val2 = out[i];
       out[i] = in[i] + val2;
  
  }
 }

MPI::Op v103;



void run() {
// Init global cartesian topology
thisRank = MPI::COMM_WORLD.Get_rank();
Nproc = MPI::COMM_WORLD.Get_size();
dimSizes = flocc::findBestDims(Nproc, ndims);

bool periods[ndims]; for (int i = 0; i < ndims; i++) periods[i] = false;
cartComm = MPI::COMM_WORLD.Create_cart(ndims, &dimSizes.front(), periods, true);

{
 int coord[ndims];
for (coord[0] = 0; coord[0] < dimSizes[0]; coord[0]++) {
coordRanks[coord[0]] = cartComm.Get_cart_rank(coord);
}

}
rootRank = coordRanks[0];
v103.Init(&userFun98, true);

  double v1;
  v1 = 0.0;
  int v2;
  v2 = 400000000;
  int v7;
  v7 = 200000000;
// BEGIN genFloatList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN genFloatList\n";
#endif
#ifdef DEBUG
std::cerr << "WARNING: genFloatList ignores par dims and distributions over ALL nodes.\n";
#endif
int v3 = v2 / Nproc + (thisRank < (v2%Nproc) ? 1 : 0);
SPTR(vector<double>) v6(new vector<double>(v3));
for (int v4 = 0; v4 < v3; v4++) {
  double v5 = (double)rand()/RAND_MAX;
  v6->push_back(v5);
}
// END genFloatList
#ifdef DEBUG
std::cerr << thisRank << ") END genFloatList\n";
#endif
// BEGIN genFloatList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN genFloatList\n";
#endif
#ifdef DEBUG
std::cerr << "WARNING: genFloatList ignores par dims and distributions over ALL nodes.\n";
#endif
int v8 = v7 / Nproc + (thisRank < (v7%Nproc) ? 1 : 0);
SPTR(vector<double>) v11(new vector<double>(v8));
for (int v9 = 0; v9 < v8; v9++) {
  double v10 = (double)rand()/RAND_MAX;
  v11->push_back(v10);
}
// END genFloatList
#ifdef DEBUG
std::cerr << thisRank << ") END genFloatList\n";
#endif
// BEGIN readList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN readList\n";
#endif
int v14 = 0;
      // BEGIN reduceList init
        double v17(v1);
          double v18;
        
      // END reduceList init
      
    std::vector<double >::iterator v15(v6->begin());
    
  v14++;
  
if (true) {
  std::vector<double >::iterator v12(v11->begin());
    std::vector<double >::iterator v13(v11->end());
  
for (; v12 != v13 && v14 > 0; v12++) {
    // BEGIN zip consume:
  if (v15 < v6->end()) {    // BEGIN mapList consume:
      double v16;
        v16 = (*v12) * (*v15);
      
      // BEGIN reduceList consume
        v18 = v17 + v16;
          v17 = v18; //(Lf:(Float )) = (Lf:(Float ))
        
      // END reduceList consume
      
    // END mapList consume
    
  }
  else if (v15 == v6->end()) v14--;
  v15++;
  // END zip consume
  
}
}
      // BEGIN reduceList fin
      if (true) {
          v18 = v17; //(Lf:(Float )) = (Lf:(Float ))
        
        cartComm.Allreduce(&v18, &v17, sizeof(double), MPI_PACKED, v103);
      }
      if (cartComm != cartComm) {
        cartComm.Bcast(&v17, sizeof(double), MPI_PACKED, rootRank);
      }
      // END reduceList fin
      
// END readList
#ifdef DEBUG
std::cerr << thisRank << ") END readList\n";
#endif
// BEGIN printFloat
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN printFloat\n";
#endif
if (thisRank == rootRank) std::cout << v17;
// END printFloat
#ifdef DEBUG
std::cerr << thisRank << ") END printFloat\n";
#endif
// No stream called 'final' found in output streams.
}


int main(int argc, char** argv) {
  MPI::Init(argc, argv);
  int rank = MPI::COMM_WORLD.Get_rank();
  double startTime = MPI::Wtime();

  run();

  double duration = MPI::Wtime() - startTime;
  //std::cerr << "\n";
  //std::cerr << rank << ") Duration " << duration << "s\n"; 
  if (rank == 0) {
    std::cerr << "dotpg\t400000000\t200000000\t" << Nproc << "\t" << duration << "\n";
  }

  MPI::Finalize();
  return 0;
}
//



